@@ -1,5 +1,5 @@
 /*
- * Ephemera Copyright (c) 2014-2025, James Bailie.
+ * Ephemera Copyright (c) 2014-2026, James Bailie.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -328,8 +328,16 @@ void become_daemon()
    else
    {
       char buffer[ 16 ];
+      int n;
+
+      n = snprintf( buffer, sizeof( buffer ), "%d", getpid() );
+
+      if ( n < 0 || n >= sizeof( buffer ))
+      {
+         syslog( LOG_ERR, "become_daemon(): buffer overflow" );
+         exit( 1 );
+      }
 
-      snprintf( buffer, sizeof( buffer ), "%d", getpid() );
       write( file, buffer, strlen( buffer ));
       close( file );
    }
@@ -805,8 +813,9 @@ void read_event( struct ccb *item )
 void add_headers( struct ccb *item )
 {
    char links[ 512 ], js[ 8192 ], buffer[ IOSIZE ], *form, *javascript, *ptr;
+   int n;
 
-   snprintf( js, sizeof( js ),
+   n = snprintf( js, sizeof( js ),
       "<script type=\"text/javascript\">\n"
       "var ephemera = function() {\n"
       "var content, callink, xhr, request = '';\n"
@@ -881,19 +890,38 @@ void add_headers( struct ccb *item )
       "</script>\n",
       path );
 
+   if ( n < 0 || n >= sizeof( js ))
+   {
+      syslog( LOG_ERR, "add_headers(): js overflow" );
+      return;
+   }
+
    if ( permalink )
    {
-      snprintf( links, sizeof( links ), "<div id=\"links\">\n<p><a href=\"%s\">New Posts</a></p>\n</div>\n", path );
+      n = snprintf( links, sizeof( links ), "<div id=\"links\">\n<p><a href=\"%s\">New Posts</a></p>\n</div>\n", path );
+
+      if ( n < 0 || n >= sizeof( links ))
+      {
+         syslog( LOG_ERR, "add_headers(): links overflow" );
+         return;
+      }
+
       javascript = form = "";
    }
    else
    {
-      snprintf( links, sizeof( links ), "%s",
-         "<div id=\"links\">\n"
-         "<p><a href=\"javascript:ephemera.load('')\">New</a> | "
-         "<a href=\"javascript:ephemera.load('request=oldest')\">Old</a> | "
-         "<a id=\"callink\" href=\"javascript:ephemera.load('request=calendar')\">Year</a>"
-         "</p></div>\n" );
+      n = snprintf( links, sizeof( links ), "%s",
+            "<div id=\"links\">\n"
+            "<p><a href=\"javascript:ephemera.load('')\">New</a> | "
+            "<a href=\"javascript:ephemera.load('request=oldest')\">Old</a> | "
+            "<a id=\"callink\" href=\"javascript:ephemera.load('request=calendar')\">Year</a>"
+            "</p></div>\n" );
+
+      if ( n < 0 || n >= sizeof( links ))
+      {
+         syslog( LOG_ERR, "add_headers(): links overflow" );
+         return;
+      }
 
       form = "<div id=\"form\">\n"
                "<form id=\"search\" name=\"search\" action=\"javascript:ephemera.search()\" method=\"GET\" enctype=\"x-www-form-urlencoded\">\n"
@@ -905,7 +933,7 @@ void add_headers( struct ccb *item )
       javascript = js;
    }
 
-   snprintf( buffer, sizeof( buffer ), "Content-Type: text/html; charset=utf-8\r\n"
+   n = snprintf( buffer, sizeof( buffer ), "Content-Type: text/html; charset=utf-8\r\n"
       "Cache-Control: no-store, no-cache, must-revalidate, max-age=0\r\n\r\n"
 
       "<!DOCTYPE html>\n"
@@ -942,6 +970,12 @@ void add_headers( struct ccb *item )
       links,
       form );
 
+   if ( n < 0 || n >= sizeof( buffer ))
+   {
+      syslog( LOG_ERR, "add_headers(): buffer overflow" );
+      return;
+   }
+
    for( ptr = buffer; *ptr; ++ptr )
       STRING_APPEND( item->response, *ptr )
 }
@@ -987,13 +1021,21 @@ struct string *add_each( struct ccb *item, sqlite3_stmt *sql, int idx, int *offs
       ++idx;
 
       if ( ! permalink )
-         snprintf( perma, sizeof( perma ),
-                   "<div class=\"permalink\">"
-                   "<p><a href=\"%s?request=one&amp;id=%s\">permalink</a></p>"
-                   "</div>\n",
-                   path, ( char *)row->values[ 3 ] );
+      {
+         i = snprintf( perma, sizeof( perma ),
+                      "<div class=\"permalink\">"
+                      "<p><a href=\"%s?request=one&amp;id=%s\">permalink</a></p>"
+                      "</div>\n",
+                      path, ( char *)row->values[ 3 ] );
 
-      snprintf( buffer, sizeof( buffer ),
+         if ( i < 0 || i >= sizeof( perma ))
+         {
+            syslog( LOG_ERR, "add_each(): perma overflow" );
+            return NULL;
+         }
+      }
+      
+      i = snprintf( buffer, sizeof( buffer ),
                 "<div class=\"article\">\n"
                 "<!-- %s -->\n"
                 "%s"
@@ -1006,6 +1048,12 @@ struct string *add_each( struct ccb *item, sqlite3_stmt *sql, int idx, int *offs
                 row->values[ 0 ],
                 row->values[ 1 ] );
 
+      if ( i < 0 || i >= sizeof( buffer ))
+      {
+         syslog( LOG_ERR, "add_each(): buffer overflow" );
+         return NULL;
+      }
+
       for( ptr = buffer; *ptr; ++ptr )
          STRING_APPEND( local, *ptr )
 
@@ -1042,7 +1090,7 @@ struct string *add_each( struct ccb *item, sqlite3_stmt *sql, int idx, int *offs
 void add_nav( struct ccb *item, sqlite3_stmt *sql, char *h2, char *request, int old_offset, int *offset, struct string *result )
 {
    char buffer[ 4096 ], before[ 1024 ], after[ 1024 ], *ptr;
-   int tmp;
+   int tmp, n;
 
    if ( permalink )
    {
@@ -1058,18 +1106,37 @@ void add_nav( struct ccb *item, sqlite3_stmt *sql, char *h2, char *request, int
    if (( tmp = old_offset - p_size ) < 0 )
       tmp = 0;
 
-   snprintf( before, sizeof( before ), "<p><a href=\"javascript:ephemera.load('request=%s&amp;index=%d')\">&lt;&lt;&lt;</a></p>", request, tmp );
-   snprintf( after, sizeof( after ), "<p><a href=\"javascript:ephemera.load('request=%s&amp;index=%d')\">&gt;&gt;&gt;</a></p>", request, *offset );
+   n = snprintf( before, sizeof( before ), "<p><a href=\"javascript:ephemera.load('request=%s&amp;index=%d')\">&lt;&lt;&lt;</a></p>", request, tmp );
 
-   snprintf( buffer, sizeof( buffer ),
-          "<div class=\"pager\">\n"
-          "<div class=\"pager_left\">%s</div>"
-          "<div class=\"pager_center\">%s</div>"
-          "<div class=\"pager_right\">%s</div>\n"
-          "</div>\n",
-          ( tmp == old_offset ? "<p>&lt;&lt;&lt;</p>" : before ),
-          h2,
-          (( result == NULL || abs(old_offset - *offset) < p_size ) ? "<p>&gt;&gt;&gt;</p>" : after ));
+   if ( n < 0 || n >= sizeof( before ))
+   {
+      syslog( LOG_ERR, "add_nav(): before overflow" );
+      return;
+   }
+   
+   n = snprintf( after, sizeof( after ), "<p><a href=\"javascript:ephemera.load('request=%s&amp;index=%d')\">&gt;&gt;&gt;</a></p>", request, *offset );
+
+   if ( n < 0 || n >= sizeof( after ))
+   {
+      syslog( LOG_ERR, "add_nav(): after overflow" );
+      return;
+   }
+
+   n = snprintf( buffer, sizeof( buffer ),
+             "<div class=\"pager\">\n"
+             "<div class=\"pager_left\">%s</div>"
+             "<div class=\"pager_center\"><p>%s</p></div>"
+             "<div class=\"pager_right\">%s</div>\n"
+             "</div>\n",
+             ( tmp == old_offset ? "<p>&lt;&lt;&lt;</p>" : before ),
+             h2,
+             (( result == NULL || abs(old_offset - *offset) < p_size ) ? "<p>&gt;&gt;&gt;</p>" : after ));
+
+   if ( n < 0 || n >= sizeof( buffer ))
+   {
+      syslog( LOG_ERR, "add_nav(): pager overflow" );
+      return;
+   }
 
    for( ptr = buffer; *ptr; ++ptr )
       STRING_APPEND( item->response, *ptr )
@@ -1098,9 +1165,15 @@ void add_results( struct ccb *item, char *h2, sqlite3_stmt *sql, char *request,
       add_nav( item, sql, h2, request, offset, &new_offset, add_each( item, sql, offset, &new_offset ));
    else
    {
-      snprintf( buffer, sizeof( buffer ),
-               "<div id=\"nomore\"><p>No %sposts.</p></div>",
-               ( offset ? "more " : "" ));
+      result = snprintf( buffer, sizeof( buffer ),
+                  "<div id=\"nomore\"><p>No %sposts.</p></div>",
+                  ( offset ? "more " : "" ));
+
+      if ( result < 0 || result >= sizeof( buffer ))
+      {
+         syslog( LOG_ERR, "add_results(): buffer overflow" );
+         return;
+      }
 
       if ( offset )
          add_nav( item, sql, h2, request, offset, &new_offset, NULL );
@@ -1183,6 +1256,8 @@ void add_one( struct ccb *item )
 
    if (( id = get_var( "id", item->params )) == NULL )
    {
+	   /* Do not remove */
+	   permalink = 0;
       add_recent( item );
       return;
    }
@@ -1200,7 +1275,7 @@ void add_one( struct ccb *item )
 void add_month( struct ccb *item )
 {
    char *t, *eom, *ptr, *month, mbuffer[ 128 ];
-   int idx;
+   int idx, n;
 
    if (( t = get_var( "time", item->params )) == NULL )
    {
@@ -1235,14 +1310,21 @@ void add_month( struct ccb *item )
       return;
    }
 
-   snprintf( mbuffer, sizeof( mbuffer ), "month&amp;time=%s", t );
+   n = snprintf( mbuffer, sizeof( mbuffer ), "month&amp;time=%s", t );
+
+   if ( n < 0 || n >= sizeof( mbuffer ))
+   {
+      syslog( LOG_ERR, "add_month(): mbuffer overflow" );
+      return;
+   }
+
    add_results( item, month, era_sql, mbuffer, idx );
 }
 
 void add_day( struct ccb *item )
 {
    char *t, *eod, *ptr, *day, dbuffer[ 128 ];
-   int idx;
+   int idx, n;
 
    if (( t = get_var( "time", item->params )) == NULL )
    {
@@ -1277,7 +1359,14 @@ void add_day( struct ccb *item )
       return;
    }
 
-   snprintf( dbuffer, sizeof( dbuffer ), "day&amp;time=%s", t );
+   n = snprintf( dbuffer, sizeof( dbuffer ), "day&amp;time=%s", t );
+
+   if ( n < 0 || n >= sizeof( dbuffer ))
+   {
+      syslog( LOG_ERR, "add_day(): dbuffer overflow" );
+      return;
+   }
+   
    add_results( item, day, era_sql, dbuffer, idx );
 }
 
@@ -1346,18 +1435,37 @@ char *compose_calendar( int this_year, int year )
    row = make_stack();
    result = make_string();
 
-   snprintf( before, sizeof( before ), "<div class=\"pager_left\"><p><a href=\"javascript:ephemera.load('request=calendar&amp;year=%d')\">&lt;&lt;&lt;</a></p></div>", year - 1 );
-   snprintf( after, sizeof( after ), "<div class=\"pager_right\"><p><a href=\"javascript:ephemera.load('request=calendar&amp;year=%d')\">&gt;&gt;&gt;</a></p></div>\n", year + 1 );
+   i = snprintf( before, sizeof( before ), "<div class=\"pager_left\"><p><a href=\"javascript:ephemera.load('request=calendar&amp;year=%d')\">&lt;&lt;&lt;</a></p></div>", year - 1 );
 
-   snprintf( header, sizeof( header ),
-             "<div class=\"pager\">\n"
-             "%s"
-             "<div class=\"pager_center\"><p>%d</p></div>"
-             "%s"
-             "</div>\n",
-             ( year < 2011 ? "<div class=\"pager_left\"><p>&lt;&lt;&lt;</p></div>" : before ),
-             year,
-             ( year == this_year ? "<div class=\"pager_right\"><p>&gt;&gt;&gt;</p></div>\n" : after ));
+   if ( i < 0 || i >= sizeof( before ))
+   {
+      syslog( LOG_ERR, "compose_calendar(): before overflow" );
+      return NULL;
+   }
+
+   i = snprintf( after, sizeof( after ), "<div class=\"pager_right\"><p><a href=\"javascript:ephemera.load('request=calendar&amp;year=%d')\">&gt;&gt;&gt;</a></p></div>\n", year + 1 );
+
+   if ( i < 0 || i >= sizeof( after ))
+   {
+      syslog( LOG_ERR, "compose_calendar(): after overflow" );
+      return NULL;
+   }
+
+   i = snprintf( header, sizeof( header ),
+                "<div class=\"pager\">\n"
+                "%s"
+                "<div class=\"pager_center\"><p>%d</p></div>"
+                "%s"
+                "</div>\n",
+                ( year < 2011 ? "<div class=\"pager_left\"><p>&lt;&lt;&lt;</p></div>" : before ),
+                year,
+                ( year == this_year ? "<div class=\"pager_right\"><p>&gt;&gt;&gt;</p></div>\n" : after ));
+
+   if ( i < 0 || i >= sizeof( header ))
+   {
+      syslog( LOG_ERR, "compose_calendar(): header overflow" );
+      return NULL;
+   }
 
    for( ptr = header; *ptr; ++ptr )
       STRING_APPEND( result, *ptr )
@@ -1502,8 +1610,8 @@ char *form_encode( char *str )
 
 void add_search( struct ccb *item )
 {
-   char *key, *ptr, request[ 256 ];
-   int idx;
+   char *key, *ptr, request[ 1024 ];
+   int idx, n;
 
    if (( key = get_var( "key", item->params )) == NULL )
    {
@@ -1512,6 +1620,13 @@ void add_search( struct ccb *item )
       return;
    }
 
+   if ( strlen( key ) > 512 )
+   {
+      syslog( LOG_ERR, "search key > 512 bytes" );
+      add_error( item );
+      return;
+   }
+
    if (( ptr = get_var( "index", item->params )) == NULL )
    {
       idx = 0;
@@ -1532,7 +1647,14 @@ void add_search( struct ccb *item )
       return;
    }
 
-   snprintf( request, sizeof( request ), "search&key=%s", form_encode( key ));
+   n = snprintf( request, sizeof( request ), "search&key=%s", form_encode( key ));
+
+   if ( n < 0 || n >= sizeof( request ))
+   {
+      syslog( LOG_ERR, "add_search(): request overflow" );
+      return;
+   }
+
    add_results( item, "Search Results", search_sql, request, idx );
 }
 
@@ -1582,6 +1704,8 @@ void set_permalink( struct ccb *item )
 {
    char *ptr1, *ptr2;
 
+   permalink = 0;
+   
    if (( ptr1 = get_var( "request", item->params )) == NULL )
       return;
 
@@ -1602,6 +1726,8 @@ int process_request( struct ccb *item )
    char *ptr;
    int xhr;
 
+   set_permalink( item );
+
    xhr = ( get_var( "xhr", item->params ) != NULL );
 
    if ( xhr )
@@ -1609,21 +1735,14 @@ int process_request( struct ccb *item )
                   "Cache-Control: no-store, must-revalidate, max-age=0\r\n\r\n"; *ptr; ++ptr )
          STRING_APPEND( item->response, *ptr )
    else
-   {
-      set_permalink( item );
       add_headers( item );
-   }
 
    add_response( item );
 
    if ( ! xhr )
-   {
       for( ptr = "</div>\n</div>\n</body>\n</html>\n"; *ptr; ++ptr )
          STRING_APPEND( item->response, *ptr )
 
-      permalink = 0;
-   }
-
    return item->response->used;
 }
 
@@ -1753,6 +1872,7 @@ void open_db()
    char buffer[ PATH_MAX ];
    sqlite3 *ddb;
    sqlite3_backup *backup;
+   int n;
 
    reload = 0;
 
@@ -1768,7 +1888,13 @@ void open_db()
       }
    }
 
-   snprintf( buffer, sizeof( buffer ), "%s/ephemera.sqlite", directory );
+   n = snprintf( buffer, sizeof( buffer ), "%s/ephemera.sqlite", directory );
+
+   if ( n < 0 || n >= sizeof( buffer ))
+   {
+      syslog( LOG_ERR, "open_db(): buffer overflow" );
+      return;
+   }
 
    if ( ! in_memory )
    {
@@ -1779,16 +1905,16 @@ void open_db()
    {
       if (( ddb = sqlite_open( buffer )) == NULL )
          exit( 1 );
-   
+
       if (( db = sqlite_open( ":memory:" )) == NULL )
          exit( 1 );
-      
+
       if (( backup = sqlite3_backup_init( db, "main", ddb, "main" )) == NULL )
       {
          syslog( LOG_ERR, "sqlite3_backup_init(): %s", ( char *)sqlite3_errmsg( db ));
          exit( 1 );
       }
-   
+
       sqlite3_backup_step( backup, -1 );
       sqlite3_backup_finish( backup );
       sqlite3_close( ddb );
@@ -1814,8 +1940,15 @@ void create_db()
 {
    char buffer[ PATH_MAX ];
    struct stat st;
+   int n;
 
-   snprintf( buffer, sizeof( buffer ), "%s/ephemera.sqlite", directory );
+   n = snprintf( buffer, sizeof( buffer ), "%s/ephemera.sqlite", directory );
+
+   if ( n < 0 || n >= sizeof( buffer ))
+   {
+      syslog( LOG_ERR, "create_db(): buffer overflow" );
+      return;
+   }
 
    if ( stat( buffer, &st ) ==  0 )
       return;
@@ -1887,7 +2020,7 @@ void process_clients()
       exit( 1 );
 
    bzero( inqueue, sizeof( struct kevent ) * qlen );
-   
+
    if (( outqueue = memory( sizeof( struct kevent ) * qlen )) == NULL )
       exit( 1 );
 
@@ -2003,39 +2136,95 @@ void set_variable( char *variable, char *value )
 
 void make_static()
 {
-   snprintf( css, sizeof( css ),
-      "<link rel=\"stylesheet\" media=\"screen\" href=\"%sephemera.css\" type=\"text/css\">\n",
-      web_directory );
+   int n;
+
+   n = snprintf( css, sizeof( css ),
+         "<link rel=\"stylesheet\" media=\"screen\" href=\"%sephemera.css\" type=\"text/css\">\n",
+         web_directory );
+
+   if ( n < 0 || n >= sizeof( css ))
+   {
+      syslog( LOG_ERR, "make_static(): css overflow" );
+      return;
+   }
 
    if ( meta_description == NULL )
       *desc = '\0';
    else
-      snprintf( desc, sizeof( desc ), "<meta name=\"description\" content=\"%s\">\n", meta_description );
+   {
+      n = snprintf( desc, sizeof( desc ), "<meta name=\"description\" content=\"%s\">\n", meta_description );
+
+      if ( n < 0 || n >= sizeof( desc ))
+      {
+         syslog( LOG_ERR, "make_static(): desc overflow" );
+         return;
+      }
+   }
 
    if ( meta_keywords == NULL )
       *keys = '\0';
    else
-      snprintf( keys, sizeof( keys ), "<meta name=\"keywords\" content=\"%s\">\n", meta_keywords );
+   {
+      n = snprintf( keys, sizeof( keys ), "<meta name=\"keywords\" content=\"%s\">\n", meta_keywords );
+
+      if ( n < 0 || n >= sizeof( keys ))
+      {
+         syslog( LOG_ERR, "make_static(): keys overflow" );
+         return;
+      }
+   }
 
    if ( meta_language == NULL )
       *lang = '\0';
    else
-      snprintf( lang, sizeof( lang ), "<meta name=\"language\" content=\"%s\">\n", meta_language );
+   {
+      n = snprintf( lang, sizeof( lang ), "<meta name=\"language\" content=\"%s\">\n", meta_language );
+
+      if ( n < 0 || n >= sizeof( lang ))
+      {
+         syslog( LOG_ERR, "make_static(): lang overflow" );
+         return;
+      }
+   }
 
    if ( meta_robots == NULL )
       *robs = '\0';
    else
-      snprintf( robs, sizeof( robs ), "<meta name=\"robots\" content=\"%s\">\n", meta_robots );
+   {
+      n = snprintf( robs, sizeof( robs ), "<meta name=\"robots\" content=\"%s\">\n", meta_robots );
+
+      if ( n < 0 || n >= sizeof( robs ))
+      {
+         syslog( LOG_ERR, "make_static(): robs overflow" );
+         return;
+      }
+   }
 
    if ( favicon == NULL )
       *fav = '\0';
    else
-      snprintf( fav, sizeof( fav ), "<link rel=\"shortcut icon\" href=\"%s%s\">\n", web_directory, favicon );
+   {
+      n = snprintf( fav, sizeof( fav ), "<link rel=\"shortcut icon\" href=\"%s%s\">\n", web_directory, favicon );
+
+      if ( n < 0 || n >= sizeof( fav ))
+      {
+         syslog( LOG_ERR, "make_static(): fav overflow" );
+         return;
+      }
+   }
 
    if ( blurb == NULL )
       *blb = '\0';
    else
-      snprintf( blb, sizeof( blb ), "<div id=\"blurb\"><p>%s</p></div>\n", blurb );
+   {
+      n = snprintf( blb, sizeof( blb ), "<div id=\"blurb\"><p>%s</p></div>\n", blurb );
+
+      if ( n < 0 || n >= sizeof( blb ))
+      {
+         syslog( LOG_ERR, "make_static(): blb overflow" );
+         return;
+      }
+   }
 }
 
 void check_vars()
@@ -2101,7 +2290,13 @@ void set_config()
    char buffer[ PATH_MAX ];
    int line;
 
-   snprintf( buffer, sizeof( buffer ), "%s/ephemera.config", config_dir );
+   line = snprintf( buffer, sizeof( buffer ), "%s/ephemera.config", config_dir );
+
+   if ( line < 0 || line >= sizeof( buffer ))
+   {
+      syslog( LOG_ERR, "set_config(): buffer overflow" );
+      exit( 1 );
+   }
 
    if (( config = fopen( buffer, "r" )) == NULL )
    {
